class CollectionNotFoundError(Exception):
    def __init__(self, collection_id=None, collection_name=None):
        if collection_id:
            super().__init__(f"Collection with id {collection_id} not found")
        elif collection_name:
            super().__init__(f"Collection with name {collection_name} not found")
        else:
            super().__init__("Collection not found")

class BookNotFoundError(Exception):
    def __init__(self, book_id=None, asset_id=None, title=None, author=None, genre=None, collection_id=None):
        if book_id:
            super().__init__(f"Book with id {book_id} not found")
        elif asset_id:
            super().__init__(f"Book with asset id {asset_id} not found")
        elif title:
            super().__init__(f"Book with title {title} not found")
        elif author:
            super().__init__(f"Book with author {author} not found")
        elif genre:
            super().__init__(f"Book with genre {genre} not found")
        elif collection_id:
            super().__init__(f"Book with collection id {collection_id} not found")
        else:
            super().__init__("Book not found")

class AnnotationNotFoundError(Exception):
    def __init__(self, annotation_id=None, asset_id=None, representative_text=None, selected_text=None, note=None, chapter=None, location=None):
        if annotation_id:
            super().__init__(f"Annotation with id {annotation_id} not found")
        elif asset_id:
            super().__init__(f"Annotation with asset id {asset_id} not found")
        elif representative_text:
            super().__init__(f"Annotation with representative text {representative_text} not found")
        elif selected_text:
            super().__init__(f"Annotation with selected text {selected_text} not found")
        elif note:
            super().__init__(f"Annotation with note {note} not found")
        elif chapter:
            super().__init__(f"Annotation with chapter {chapter} not found")
        elif location:
            super().__init__(f"Annotation with location {location} not found")
        else:
            super().__init__("Annotation not found")


from functools import lru_cache
from py_apple_books.utils import get_mappings

COLLECTION_TABLE_NAME = "ZBKCOLLECTION"
BOOK_TABLE_NAME = "ZBKLIBRARYASSET"
ANNOTATION_TABLE_NAME = "anno_db.ZAEANNOTATION"
COLLECTIONMEMBER_TABLE_NAME = "ZBKCOLLECTIONMEMBER"

@lru_cache(maxsize=5)
def get_fields_str(class_name: str, table_name: str) -> str:
    mappings = get_mappings(class_name)
    return ", ".join(f"{table_name}.{mappings[field]} AS {field}" for field in mappings)

@lru_cache(maxsize=1)
def get_book_annotation_query() -> str:
    book_fields_str = get_fields_str('Book', BOOK_TABLE_NAME)
    annotation_fields_str = get_fields_str('Annotation', ANNOTATION_TABLE_NAME)
    query = f"""
        SELECT {book_fields_str}, {annotation_fields_str}
        FROM {BOOK_TABLE_NAME}
        LEFT JOIN {ANNOTATION_TABLE_NAME}
        ON {BOOK_TABLE_NAME}.ZASSETID = {ANNOTATION_TABLE_NAME}.ZANNOTATIONASSETID
    """
    return query

@lru_cache(maxsize=1)
def book_collection_member_query() -> str:
    fields_str = get_fields_str('Book', BOOK_TABLE_NAME)
    query = f"""
        SELECT {fields_str}
        FROM {BOOK_TABLE_NAME}
        INNER JOIN {COLLECTIONMEMBER_TABLE_NAME}
        ON {COLLECTIONMEMBER_TABLE_NAME}.ZASSET = {BOOK_TABLE_NAME}.Z_PK
    """
    return query

@lru_cache(maxsize=1)
def get_book_collection_query() -> str:
    collection_fields_str = get_fields_str('Collection', COLLECTION_TABLE_NAME)
    book_fields_str = get_fields_str('Book', BOOK_TABLE_NAME)
    query = f"""
        SELECT {collection_fields_str}, {book_fields_str}
        FROM {COLLECTION_TABLE_NAME}
        JOIN {COLLECTIONMEMBER_TABLE_NAME}
        ON {COLLECTION_TABLE_NAME}.Z_PK = {COLLECTIONMEMBER_TABLE_NAME}.ZCOLLECTION
        JOIN {BOOK_TABLE_NAME}
        ON {COLLECTIONMEMBER_TABLE_NAME}.ZASSETID = {BOOK_TABLE_NAME}.ZASSETID
    """
    return query

@lru_cache(maxsize=1)
def get_collection_collection_member_query() -> str:
    collection_fields_str = get_fields_str('Collection', COLLECTION_TABLE_NAME)
    query = f"""
        SELECT {collection_fields_str}
        FROM {COLLECTION_TABLE_NAME}
        JOIN {COLLECTIONMEMBER_TABLE_NAME}
        ON {COLLECTION_TABLE_NAME}.Z_PK = {COLLECTIONMEMBER_TABLE_NAME}.ZCOLLECTION
    """
    return query



from dataclasses import dataclass
from datetime import datetime

@dataclass
class Annotation:
    # Identifiers
    id: str
    asset_id: str

    # Status
    is_deleted: bool

    # Dates
    creation_date: datetime
    modification_date: datetime

    # Annotation details
    representative_text: str
    selected_text: str
    note: str

    # Location
    chapter: str
    location: str

    def __post_init__(self):
        """
        Converts the creation_date and modification_date from timestamp to datetime.
        """
        self.creation_date = datetime.fromtimestamp(float(self.creation_date) / 1000) if self.creation_date else None
        self.modification_date = datetime.fromtimestamp(float(self.modification_date) / 1000) if self.modification_date else None

    def __hash__(self):
        """
        Returns a hash value for the Annotation object.

        Returns:
            int: hash value for the Annotation object
        """
        return hash(self.id)


import sqlite3
from pathlib import Path
from py_apple_books.data import db_utils, query_utils
from py_apple_books.utils import get_mappings
from functools import lru_cache

def find_all() -> list:
    """
    Fetches all annotations in the database

    Returns:
        list: list of all annotations in the database
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    return db_utils.find_all(fields_str, query_utils.ANNOTATION_TABLE_NAME)

def find_by_id(annotation_id: str) -> list:
    """
    Fetches an annotation by its id

    Args:
        annotation_id (str): id of the annotation

    Returns:
        list: list of annotations with the specified id
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    return db_utils.find_by_field(fields_str, query_utils.ANNOTATION_TABLE_NAME, "Z_PK", annotation_id)

def find_by_book_id(book_id: str) -> list:
    """
    Fetches all annotations for a book

    Args:
        book_id (str): id of the book

    Returns:
        list: list of annotations for the specified book
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    return db_utils.find_by_field(fields_str, query_utils.ANNOTATION_TABLE_NAME, "ZANNOTATIONASSETID", book_id)

def find_deleted() -> list:
    """
    Fetches all deleted annotations

    Returns:
        list: list of deleted annotations
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    return db_utils.find_by_field(fields_str, query_utils.ANNOTATION_TABLE_NAME, "ZANNOTATIONDELETED", 1)

def find_not_deleted() -> list:
    """
    Fetches all non-deleted annotations

    Returns:
        list: list of non-deleted annotations
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    return db_utils.find_by_field(fields_str, query_utils.ANNOTATION_TABLE_NAME, "ZANNOTATIONDELETED", 0)

def find_underlines() -> list:
    """
    Fetches all underlines

    Returns:
        list: list of underlines
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    return db_utils.find_by_field(fields_str, query_utils.ANNOTATION_TABLE_NAME, "ZANNOTATIONISUNDERLINE", 1)

def find_by_chapter(chapter: str) -> list:
    """
    Fetches all annotations for a chapter

    Args:
        chapter (str): chapter number

    Returns:
        list: list of annotations for the specified chapter
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    return db_utils.find_by_field(fields_str, query_utils.ANNOTATION_TABLE_NAME, "ZFUTUREPROOFING5", chapter)

def find_by_location(location: str) -> list:
    """
    Fetches annotation for a location

    Args:
        location (str): location of the annotation

    Returns:
        list: list of annotations for the specified location
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    return db_utils.find_by_field(fields_str, query_utils.ANNOTATION_TABLE_NAME, "ZANNOTATIONLOCATION", location)

def find_by_representative_text(representative_text: str) -> list:
    """
    Fetches annotations by representative text

    Args:
        representative_text (str): representative text of the annotation

    Returns:
        list: list of annotations with the specified representative text
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    query = f"""
        SELECT {fields_str}
        FROM {query_utils.ANNOTATION_TABLE_NAME}
        WHERE ZANNOTATIONREPRESENTATIVETEXT LIKE \'{representative_text}\'
    """
    return db_utils.run_query(query)

def find_by_note_text(note_text: str) -> list:
    """
    Fetches annotations by note text

    Args:
        note_text (str): note text of the annotation

    Returns:
        list: list of annotations with the specified note text
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    query = f"""
        SELECT {fields_str}
        FROM {query_utils.ANNOTATION_TABLE_NAME}
        WHERE ZANNOTATIONNOTE LIKE \'{note_text}\'
    """
    return db_utils.run_query(query)

def find_by_selected_text(selected_text: str):
    """
    Fetches annotations by selected text

    Args:
        selected_text (str): selected text of the annotation

    Returns:
        list: list of annotations with the specified selected text
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    query = f"""
        SELECT {fields_str}
        FROM {query_utils.ANNOTATION_TABLE_NAME}
        WHERE ZANNOTATIONSELECTEDTEXT LIKE \'{selected_text}\'
    """
    return db_utils.run_query(query)

def find_highlights() -> list:
    """
    Fetches all highlights

    Returns:
        list: list of highlights
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    query = f"""
        SELECT {fields_str}
        FROM {query_utils.ANNOTATION_TABLE_NAME}
        WHERE ZANNOTATIONISUNDERLINE = 0 AND ZANNOTATIONSTYLE != 0
    """
    return db_utils.run_query(query)

def find_notes() -> list:
    """
    Fetches all notes

    Returns:
        list: list of notes
    """
    fields_str = query_utils.get_fields_str('Annotation', query_utils.ANNOTATION_TABLE_NAME)
    query = f"""
        SELECT {fields_str}
        FROM {query_utils.ANNOTATION_TABLE_NAME}
        WHERE ZANNOTATIONNOTE != ''
    """
    return db_utils.run_query(query)


from functools import lru_cache
from pathlib import Path
import sqlite3
from py_apple_books.utils import get_mappings

DB_PATHS = [
    (Path.home() / "Library/Containers/com.apple.iBooksX/Data/Documents/BKLibrary"),
    (Path.home() / "Library/Containers/com.apple.iBooksX/Data/Documents/AEAnnotation")
]

@lru_cache(maxsize=1)
def get_db_cursor() -> sqlite3.Cursor:
    """
    Gets a cursor for the database

    Returns:
        sqlite3.Cursor: cursor for the database
    """
    sqlite_file = list(DB_PATHS[0].glob("*.sqlite"))[0]
    conn = sqlite3.connect(sqlite_file)
    cursor = conn.cursor()

    other_sqlite_file = list(DB_PATHS[1].glob("*.sqlite"))[0]
    cursor.execute(f"ATTACH DATABASE '{other_sqlite_file}' AS anno_db")
    return cursor


def find_all(fields_str: str, table: str) -> list:
    """
    Fetches all rows in a table

    Args:
        fields_str (str): fields to fetch
        table (str): table to fetch from

    Returns:
        list: list of all rows in the table
    """
    cursor = get_db_cursor()
    query = f"""
        SELECT {fields_str}
        FROM {table}
    """
    cursor.execute(query)
    return cursor.fetchall()

def find_by_field(fields_str: str, table: str, field: str, value: str) -> list:
    """
    Fetches rows in a table by a field

    Args:
        fields_str (str): fields to fetch
        table (str): table to fetch from
        field (str): field to match
        value (str): value to match

    Returns:
        list: list of rows in the table matching the field and value
    """
    cursor = get_db_cursor()
    query = f"""
        SELECT {fields_str}
        FROM {table}
        WHERE {field} = ?
    """
    cursor.execute(query, (value,))
    return cursor.fetchall()

def run_query(query: str) -> list:
    cursor = get_db_cursor()
    cursor.execute(query)
    return cursor.fetchall()


from py_apple_books.models.annotation import Annotation
from py_apple_books.models.highlight import Highlight
from py_apple_books.models.underline import Underline

from dataclasses import dataclass, field
from datetime import datetime
from typing import List
import pathlib


@dataclass
class Book:
    # Identifiers
    id: str
    asset_id: str

    # Basic book information
    title: str
    author: str
    description: str
    genre: str
    content_type: str
    page_count: int

    # File information
    path: pathlib.Path
    filesize: int

    # Reading progress
    is_finished: bool
    reading_progress: float
    duration: float

    # Dates
    creation_date: datetime
    finished_date: datetime
    last_opened_date: datetime
    purchased_date: datetime

    # Flags
    is_explicit: bool
    is_locked: bool
    is_ephemeral: bool
    is_hidden: bool
    is_sample: bool
    is_store_audiobook: bool

    # User interactions
    rating: int
    highlights: List[Highlight] = field(default_factory=list)
    underlines: List[Underline] = field(default_factory=list)
    other_annotations: List[Annotation] = field(default_factory=list)

    def __post_init__(self):
        """
        Converts the creation_date, finished_date, last_opened_date, and purchased_date from timestamp to datetime.
        """
        self.creation_date = datetime.fromtimestamp(float(self.creation_date) / 1000) if self.creation_date else None
        self.finished_date = datetime.fromtimestamp(float(self.finished_date) / 1000) if self.finished_date else None
        self.last_opened_date = datetime.fromtimestamp(float(self.last_opened_date) / 1000) if self.last_opened_date else None
        self.purchased_date = datetime.fromtimestamp(float(self.purchased_date) / 1000) if self.purchased_date else None
        self.duration = float(self.duration) / 1000 if self.duration else None
        self.reading_progress = float(self.reading_progress) * 100 if self.reading_progress else None

    def __hash__(self) -> int:
        """
        Returns a hash value for the Book object.

        Returns:
            int: hash value for the Book object
        """
        return hash(self.id)

    def add_annotation(self, annotation: Annotation) -> None:
        """
        Adds an annotation to the book.

        Args:
            annotation (Annotation): annotation to add
        """
        if isinstance(annotation, Highlight):
            self.highlights.append(annotation)
        elif isinstance(annotation, Underline):
            self.underlines.append(annotation)
        else:
            self.other_annotations.append(annotation)


from py_apple_books.data import collection_db, book_db, annotation_db
from py_apple_books.models.book import Book
from py_apple_books.models.collection import Collection
from py_apple_books.models.highlight import Highlight
from py_apple_books.models.underline import Underline
from py_apple_books.models.annotation import Annotation
from py_apple_books.exceptions import CollectionNotFoundError, BookNotFoundError, AnnotationNotFoundError
from py_apple_books.utils import get_mappings
from typing import Optional


class PyAppleBooks:

    def __init__(self):
        self._book_mappings = get_mappings('Book')
        self._collection_mappings = get_mappings('Collection')
        self._annotation_mappings = get_mappings('Annotation')

        self._style_mappings = {
            1: 'Green',
            2: 'Blue',
            3: 'Yellow',
            4: 'Pink',
            5: 'Purple',
        }

        self._book_map = {}
        self._collection_map = {}

    def _get_or_create_book(self, book_params: dict) -> Book:
        """
        Returns a Book object if it exists in the book map, otherwise creates a new Book object and adds it to the book map.

        Args:
            book_params (dict): dictionary of book parameters
            
        Returns:
            Book: Book object
        """
        id_ = book_params['id']
        if id_ not in self._book_map:
            self._book_map[id_] = Book(**book_params)
        return self._book_map[id_]

    def _get_or_create_collection(self, collection_params: dict) -> Collection:
        """
        Returns a Collection object if it exists in the collection map, otherwise creates a new Collection object and adds it to the collection map.

        Args:
            collection_params (dict): dictionary of collection parameters

        Returns:    
            Collection: Collection object
        """
        id_ = collection_params['id']
        if id_ not in self._collection_map:
            self._collection_map[id_] = Collection(**collection_params)
        return self._collection_map[id_]

    def _create_annotation_object(self, annotation: dict) -> Annotation:
        """
        Returns an Annotation object.

        Args:
            annotation (dict): dictionary of annotation parameters

        Returns:
            Annotation: Annotation object
        """
        is_underline = annotation.pop('is_underline')
        style = annotation.pop('style')
        if is_underline:
            return Underline(**annotation)
        elif style in self._style_mappings:
            annotation['color'] = self._style_mappings[style]
            return Highlight(**annotation)
        else:
            return Annotation(**annotation)

    def _create_book_object(self, raw_book_data: list) -> Book:
        """
        Returns a Book object.

        Args:
            raw_book_data (list): list of book data

        Returns:
            Book: Book object
        """
        book_fields_len = len(self._book_mappings)
        book_dict = dict(zip(self._book_mappings.keys(), raw_book_data[:book_fields_len]))
        book = self._get_or_create_book(book_dict)
        if raw_book_data[book_fields_len:]:
            annotation_dict = dict(zip(self._annotation_mappings.keys(), raw_book_data[book_fields_len:]))
            if annotation_dict['id']:
                book.add_annotation(self._create_annotation_object(annotation_dict))
        return book

    def _create_collection_object(self, raw_collection_data: list) -> Collection:
        """
        Returns a Collection object.

        Args:
            raw_collection_data (list): list of collection data

        Returns:
            Collection: Collection object
        """
        collection_fields_len = len(self._collection_mappings)
        collection_dict = dict(zip(self._collection_mappings.keys(), raw_collection_data[:collection_fields_len]))
        collection = self._get_or_create_collection(collection_dict)
        if raw_collection_data[collection_fields_len:]:
            book_dict = dict(zip(self._book_mappings.keys(), raw_collection_data[collection_fields_len:]))
            if book_dict['id']:
                collection.add_book(Book(**book_dict))
        return collection

    def list_collections(self, include_books: bool = False) -> list[Collection]:
        """
        Returns a list of Collection objects.

        Args:
            include_books (bool, optional): if True, includes books in the collection. Defaults to False.

        Returns:
            list[Collection]: list of Collection objects
        """
        list_of_collections = set()
        raw_collection_data = collection_db.find_all(include_books)
        if not raw_collection_data: return []
        for collection_data in raw_collection_data:
            list_of_collections.add(self._create_collection_object(collection_data))
        return list(list_of_collections)

    def get_collection_by_id(self, collection_id: str, include_books: bool = False) -> Optional[Collection]:
        """
        Fetches a Collection object by id.

        Args:
            collection_id (str): collection id
            include_books (bool, optional): if True, includes books in the collection. Defaults to False.

        Returns:
            Collection: Collection object
        """
        raw_collection_data = collection_db.find_by_id(collection_id)
        if not raw_collection_data:
            raise CollectionNotFoundError(collection_id=collection_id)
        raw_collection_data = raw_collection_data[0]
        return self._create_collection_object(raw_collection_data)

    def get_collection_by_name(self, collection_name: str, include_books: bool = False) -> Optional[Collection]:
        """
        Fetches a Collection object by name.

        Args:
            collection_name (str): collection name
            include_books (bool, optional): if True, includes books in the collection. Defaults to False.

        Returns:
            Collection: Collection object
        """
        raw_collection_data = collection_db.find_by_name(collection_name)
        if not raw_collection_data:
            raise CollectionNotFoundError(collection_name=collection_name)
        raw_collection_data = raw_collection_data[0]
        return self._create_collection_object(raw_collection_data)

    def list_books(self, include_annotations=False) -> list[Book]:
        """
        Returns a list of Book objects.

        Args:
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            list[Book]: list of Book objects
        """
        list_of_books = set()
        raw_book_data = book_db.find_all(include_annotations)
        if not raw_book_data: return []
        for book_data in raw_book_data:
            list_of_books.add(self._create_book_object(book_data))
        return list(list_of_books)

    def get_book_by_id(self, book_id: str, include_annotations=False) -> Optional[Book]:
        """
        Fetches a Book object by id.

        Args:
            book_id (str): book id
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            Book: Book object
        """
        raw_book_data = book_db.find_by_id(book_id, include_annotations)
        if not raw_book_data:
            raise BookNotFoundError(book_id=book_id)
        raw_book_data = raw_book_data[0]
        return self._create_book_object(raw_book_data)

    def get_book_by_asset_id(self, asset_id: str, include_annotations: bool = False) -> Optional[Book]:
        """
        Fetches a Book object by asset id.

        Args:
            asset_id (str): asset id
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            Book: Book object
        """
        raw_book_data = book_db.find_by_asset_id(asset_id, include_annotations)
        if not raw_book_data:
            raise BookNotFoundError(asset_id=asset_id)
        raw_book_data = raw_book_data[0]
        return self._create_book_object(raw_book_data)

    def get_book_by_title(self, title: str, include_annotations: bool = False) -> Optional[Book]:
        """
        Fetches a Book object by title.

        Args:
            title (str): book title
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            Book: Book object
        """
        raw_book_data = book_db.find_by_title(title, include_annotations)
        if not raw_book_data:
            raise BookNotFoundError(title=title)
        raw_book_data = raw_book_data[0]
        return self._create_book_object(raw_book_data)

    def get_books_by_author(self, author: str, include_annotations: bool = False) -> Optional[Book]:
        """
        Fetches a list of Book objects by author.

        Args:
            author (str): book author
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            list[Book]: list of Book objects
        """
        list_of_books = set()
        raw_book_data = book_db.find_by_author(author, include_annotations)
        if not raw_book_data:
            raise BookNotFoundError(author=author)
        for book_data in raw_book_data:
            list_of_books.add(self._create_book_object(book_data))
        return list(list_of_books)

    def get_books_by_genre(self, genre: str, include_annotations: bool = False) -> Optional[Book]:
        """
        Fetches a list of Book objects by genre.

        Args:
            genre (str): book genre
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            list[Book]: list of Book objects
        """
        list_of_books = set()
        raw_book_data = book_db.find_by_genre(genre)
        if not raw_book_data:
            raise BookNotFoundError(genre=genre)
        for book_data in raw_book_data:
            list_of_books.add(self._create_book_object(book_data))
        return list(list_of_books)

    def get_finished_books(self, include_annotations: bool = False) -> list[Book]:
        """
        Fetches a list of finished Book objects.

        Args:
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            list[Book]: list of Book objects
        """
        list_of_books = set()
        raw_book_data = book_db.find_finished_books()
        if not raw_book_data: return []
        for book_data in raw_book_data:
            list_of_books.add(self._create_book_object(book_data))
        return list(list_of_books)

    def get_unfinished_books(self, include_annotations: bool = False) -> list[Book]:
        """
        Fetches a list of unfinished Book objects.

        Args:
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            list[Book]: list of Book objects
        """
        list_of_books = set()
        raw_book_data = book_db.find_unfinished_books()
        if not raw_book_data: return []
        for book_data in raw_book_data:
            list_of_books.add(self._create_book_object(book_data))
        return list(list_of_books)

    def get_explicit_books(self, include_annotations: bool = False) -> list[Book]:
        """
        Fetches a list of explicit Book objects.

        Args:
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            list[Book]: list of Book objects
        """
        list_of_books = set()
        raw_book_data = book_db.find_explicit_books()
        if not raw_book_data: return []
        for book_data in raw_book_data:
            list_of_books.add(self._create_book_object(book_data))
        return list(list_of_books)

    def get_locked_books(self, include_annotations: bool = False) -> list[Book]:
        """
        Fetches a list of locked Book objects.

        Args:
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            list[Book]: list of Book objects
        """
        list_of_books = set()
        raw_book_data = book_db.find_locked_books()
        if not raw_book_data: return []
        for book_data in raw_book_data:
            list_of_books.add(self._create_book_object(book_data))
        return list(list_of_books)

    def get_ephemeral_books(self, include_annotations: bool = False) -> list[Book]:
        """
        Fetches a list of ephemeral Book objects.

        Args:
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            list[Book]: list of Book objects
        """
        list_of_books = set()
        raw_book_data = book_db.find_ephemeral_books()
        if not raw_book_data: return []
        for book_data in raw_book_data:
            list_of_books.add(self._create_book_object(book_data))
        return list(list_of_books)

    def get_hidden_books(self, include_annotations: bool = False) -> list[Book]:
        """
        Fetches a list of hidden Book objects.

        Args:
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            list[Book]: list of Book objects
        """
        list_of_books = set()
        raw_book_data = book_db.find_hidden_books()
        if not raw_book_data: return []
        for book_data in raw_book_data:
            list_of_books.add(self._create_book_object(book_data))
        return list(list_of_books)

    def get_sample_books(self, include_annotations: bool = False) -> list[Book]:
        """
        Fetches a list of sample Book objects.

        Args:
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            list[Book]: list of Book objects
        """
        list_of_books = set()
        raw_book_data = book_db.find_sample_books()
        if not raw_book_data: return []
        for book_data in raw_book_data:
            list_of_books.add(self._create_book_object(book_data))
        return list(list_of_books)

    def get_books_by_rating(self, rating: int, include_annotations: bool = False) -> list[Book]:
        """
        Fetches a list of Book objects by rating.

        Args:
            rating (int): book rating
            include_annotations (bool, optional): if True, includes annotations in the book. Defaults to False.

        Returns:
            list[Book]: list of Book objects
        """
        list_of_books = set()
        raw_book_data = book_db.find_by_rating(rating)
        if not raw_book_data: return []
        for book_data in raw_book_data:
            list_of_books.add(self._create_book_object(book_data))
        return list(list_of_books)

    def get_books_by_collection_id(self, collection_id: str) -> list[Book]:
        """
        Fetches a list of Book objects by collection id.

        Args:
            collection_id (str): collection id

        Returns:
            list[Book]: list of Book objects
        """
        list_of_books = set()
        raw_book_data = book_db.find_by_collection_id(collection_id)
        if not raw_book_data:
            raise BookNotFoundError(collection_id=collection_id)
        for book_data in raw_book_data:
            list_of_books.add(self._create_book_object(book_data))
        return list(list_of_books)

    def list_annotations(self) -> list[Annotation]:
        """
        Fetches a list of Annotation objects.

        Returns:
            list[Annotation]: list of Annotation objects
        """
        list_of_annotations = []
        raw_annotation_data = annotation_db.find_all()
        if not raw_annotation_data: return []
        annotations = [dict(zip(self._annotation_mappings.keys(), a)) for a in raw_annotation_data]
        for annotation in annotations:
            list_of_annotations.append(self._create_annotation_object(annotation))
        return list_of_annotations

    def list_highlights(self) -> list[Annotation]:
        """
        Fetches a list of Highlight objects.

        Returns:
            list[Highlight]: list of Highlight objects
        """
        list_of_annotations = []
        raw_annotation_data = annotation_db.find_highlights()
        if not raw_annotation_data: return []
        annotations = [dict(zip(self._annotation_mappings.keys(), a)) for a in raw_annotation_data]
        for annotation in annotations:
            list_of_annotations.append(self._create_annotation_object(annotation))
        return list_of_annotations

    def list_underlines(self) -> list[Annotation]:
        """
        Fetches a list of Underline objects.

        Returns:
            list[Underline]: list of Underline objects
        """
        list_of_annotations = []
        raw_annotation_data = annotation_db.find_underlines()
        if not raw_annotation_data: return []
        annotations = [dict(zip(self._annotation_mappings.keys(), a)) for a in raw_annotation_data]
        for annotation in annotations:
            list_of_annotations.append(self._create_annotation_object(annotation))
        return list_of_annotations

    def get_all_notes(self) -> list[Annotation]:
        """
        Fetches a list of Annotation objects with notes.

        Returns:
            list[Annotation]: list of Annotation objects
        """
        list_of_annotations = []
        raw_annotation_data = annotation_db.find_notes()
        if not raw_annotation_data: return []
        annotations = [dict(zip(self._annotation_mappings.keys(), a)) for a in raw_annotation_data]
        for annotation in annotations:
            list_of_annotations.append(self._create_annotation_object(annotation))
        return list_of_annotations

    def get_annotation_by_id(self, annotation_id: str) -> Optional[Annotation]:
        """
        Fetches an Annotation object by id.

        Args:
            annotation_id (str): annotation id

        Returns:
            Annotation: Annotation object
        """
        raw_annotation_data = annotation_db.find_by_id(annotation_id)
        if not raw_annotation_data:
            raise AnnotationNotFoundError(annotation_id=annotation_id)
        raw_annotation_data = raw_annotation_data[0]
        annotation = dict(zip(self._annotation_mappings.keys(), raw_annotation_data))
        return self._create_annotation_object(annotation)

    def get_annotations_by_book_id(self, asset_id: str) -> Optional[Annotation]:
        """
        Fetches Annotations by book id.

        Args:
            asset_id (str): book id

        Returns:
            list[Annotation]: list of Annotation objects
        """
        raw_annotation_data = annotation_db.find_by_book_id(asset_id)
        if not raw_annotation_data:
            raise AnnotationNotFoundError(asset_id=asset_id)
        annotations = [dict(zip(self._annotation_mappings.keys(), a)) for a in raw_annotation_data]
        list_of_annotations = []
        for annotation in annotations:
            annotation_obj = self._create_annotation_object(annotation)
            list_of_annotations.append(annotation_obj)
        return list_of_annotations

    def get_annotations_by_representative_text(self, representative_text: str) -> Optional[Annotation]:
        """
        Fetches a list of Annotation objects by representative text.

        Args:
            representative_text (str): representative text

        Returns:
            list[Annotation]: list of Annotation objects
        """
        raw_annotation_data = annotation_db.find_by_representative_text(representative_text)
        if not raw_annotation_data: return []
        list_of_annotations = []
        annotations = [dict(zip(self._annotation_mappings.keys(), a)) for a in raw_annotation_data]
        for annotation in annotations:
            annotation_obj = self._create_annotation_object(annotation)
            list_of_annotations.append(annotation_obj)
        return list_of_annotations

    def get_deleted_annotations(self) -> list[Annotation]:
        """
        Fetches a list of deleted Annotation objects.

        Returns:
            list[Annotation]: list of Annotation objects
        """
        raw_annotation_data = annotation_db.find_deleted()
        if not raw_annotation_data: return []
        list_of_annotations = []
        annotations = [dict(zip(self._annotation_mappings.keys(), a)) for a in raw_annotation_data]
        for annotation in annotations:
            annotation_obj = self._create_annotation_object(annotation)
            list_of_annotations.append(annotation_obj)
        return list_of_annotations

    def get_not_deleted_annotations(self) -> list[Annotation]:
        """
        Fetches a list of not deleted Annotation objects.

        Returns:
            list[Annotation]: list of Annotation objects
        """
        raw_annotation_data = annotation_db.find_not_deleted()
        if not raw_annotation_data: return []
        list_of_annotations = []
        annotations = [dict(zip(self._annotation_mappings.keys(), a)) for a in raw_annotation_data]
        for annotation in annotations:
            annotation_obj = self._create_annotation_object(annotation)
            list_of_annotations.append(annotation_obj)
        return list_of_annotations

    def get_annotations_by_chapter(self, chapter: str) -> list[Annotation]:
        """
        Fetches a list of Annotation objects by chapter.

        Args:
            chapter (str): chapter

        Returns:
            list[Annotation]: list of Annotation objects
        """
        raw_annotation_data = annotation_db.find_by_chapter(chapter)
        if not raw_annotation_data: return []
        list_of_annotations = []
        annotations = [dict(zip(self._annotation_mappings.keys(), a)) for a in raw_annotation_data]
        for annotation in annotations:
            annotation_obj = self._create_annotation_object(annotation)
            list_of_annotations.append(annotation_obj)
        return list_of_annotations

    def get_annotation_by_location(self, location: str) -> list[Annotation]:
        """
        Fetches an Annotation object by location.

        Args:
            location (str): location

        Returns:
            Annotation: Annotation object
        """
        raw_annotation_data = annotation_db.find_by_location(location)
        if not raw_annotation_data: raise AnnotationNotFoundError(location=location)
        raw_annotation_data = raw_annotation_data[0]
        annotation = dict(zip(self._annotation_mappings.keys(), raw_annotation_data))
        return self._create_annotation_object(annotation)

    def get_annotations_by_note(self, note: str) -> list[Annotation]:
        """
        Fetches a list of Annotation objects by note.

        Args:
            note (str): note

        Returns:
            list[Annotation]: list of Annotation objects
        """
        raw_annotation_data = annotation_db.find_by_note_text(note)
        if not raw_annotation_data: return []
        list_of_annotations = []
        annotations = [dict(zip(self._annotation_mappings.keys(), a)) for a in raw_annotation_data]
        for annotation in annotations:
            annotation_obj = self._create_annotation_object(annotation)
            list_of_annotations.append(annotation_obj)
        return list_of_annotations

    def get_annotations_by_selected_text(self, selected_text: str) -> list[Annotation]:
        """
        Fetches a list of Annotation objects by selected text.

        Args:
            selected_text (str): selected text

        Returns:
            list[Annotation]: list of Annotation objects
        """
        list_of_annotations = []
        raw_annotation_data = annotation_db.find_by_selected_text(selected_text)
        if not raw_annotation_data: return []
        annotations = [dict(zip(self._annotation_mappings.keys(), a)) for a in raw_annotation_data]
        for annotation in annotations:
            annotation_obj = self._create_annotation_object(annotation)
            list_of_annotations.append(annotation_obj)
        return list_of_annotations


import sqlite3
from pathlib import Path
from py_apple_books.data import db_utils, query_utils
from functools import lru_cache

def find_all(books: bool = False) -> list:
    """
    Fetches all collections in the database

    Returns:
        list: list of all collections in the database
    """
    if not books:
        fields_str = query_utils.get_fields_str('Collection', query_utils.COLLECTION_TABLE_NAME)
        return db_utils.find_all(fields_str, query_utils.COLLECTION_TABLE_NAME)
    return db_utils.run_query(query_utils.get_book_collection_query())

def find_by_id(collection_id: str, books: bool = False) -> list:
    """
    Fetches a collection by its id

    Args:
        collection_id (str): id of the collection

    Returns:
        list: list of collections with the specified id
    """
    if not books:
        fields_str = query_utils.get_fields_str('Collection', query_utils.COLLECTION_TABLE_NAME)
        return db_utils.find_by_field(fields_str, query_utils.COLLECTION_TABLE_NAME, "Z_PK", collection_id)
    query = query_utils.get_book_collection_query() + f"WHERE {query_utils.COLLECTION_TABLE_NAME}.Z_PK = {collection_id}"
    return db_utils.run_query(query)

def find_by_name(collection_name: str, books: bool = False) -> list:
    """
    Fetches a collection by its name

    Args:
        collection_name (str): name of the collection

    Returns:
        list: list of collections with the specified name
    """
    if not books:
        fields_str = query_utils.get_fields_str('Collection', query_utils.COLLECTION_TABLE_NAME)
        return db_utils.find_by_field(fields_str, query_utils.COLLECTION_TABLE_NAME, "ZTITLE", collection_name)
    query = query_utils.get_book_collection_query() + f"WHERE {query_utils.COLLECTION_TABLE_NAME}.ZTITLE = {collection_name}"
    return db_utils.run_query(query)

def find_by_book_id(book_id: str) -> list:
    """
    Fetches all collections for a book

    Args:
        book_id (str): id of the book

    Returns:
        list: list of collections for the specified book
    """
    query = query_utils.get_book_collection_query() + f"WHERE {query_utils.BOOK_TABLE_NAME}.Z_PK = {book_id}"
    return db_utils.run_query(query)


import sqlite3
from pathlib import Path
from py_apple_books.data import db_utils, query_utils
from functools import lru_cache

def find_all(annotations: bool = False) -> list:
    """
    Fetches all books in the database

    Returns:
        list: list of all books in the database
    """
    if not annotations:
        fields_str = db_utils.get_fields_str('Book', query_utils.BOOK_TABLE_NAME)
        return db_utils.find_all(fields_str, BOOK_TABLE_NAME)
    return db_utils.run_query(query_utils.get_book_annotation_query())

def find_by_id(book_id: str, annotations: bool = False) -> list:
    """
    Fetches a book by its id

    Args:
        book_id (str): id of the book

    Returns:
        list: list of books with the specified id
    """
    if not annotations:
        fields_str = query_utils.get_fields_str('Book', query_utils.BOOK_TABLE_NAME)
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "Z_PK", book_id)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.Z_PK = {book_id}"
    return db_utils.run_query(query)

def find_by_asset_id(asset_id: str, annotations: bool = False) -> list:
    """
    Fetches a book by its asset id

    Args:
        asset_id (str): asset id of the book

    Returns:
        list: list of books with the specified asset id
    """
    if not annotations:
        fields_str = query_utils.get_fields_str('Book', query_utils.BOOK_TABLE_NAME)
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZASSETID", asset_id)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZASSETID = \'{asset_id}\'"
    return db_utils.run_query(query)

def find_by_title(title: str, annotations: bool = False) -> list:
    """
    Fetches a book by its title

    Args:
        title (str): title of the book

    Returns:
        list: list of books with the specified title
    """
    if not annotations:
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZTITLE", title)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZTITLE = '{title}'"
    print(query)
    return db_utils.run_query(query)

def find_by_author(author: str, annotations: bool = False) -> list:
    """
    Fetches a book by its author

    Args:
        author (str): author of the book

    Returns:
        list: list of books with the specified author
    """
    if not annotations:
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZAUTHOR", author)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZAUTHOR = '{author}'"
    return db_utils.run_query(query)

def find_by_genre(genre: str, annotations: bool = False) -> list:
    """
    Fetches a book by its genre
    
    Args:
        genre (str): genre of the book

    Returns:
        list: list of books with the specified genre
    """
    if not annotations:
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZGENRE", genre)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZGENRE = '{genre}'"
    return db_utils.run_query(query)

def find_finished_books(annotations: bool = False) -> list:
    """
    Fetches all finished books

    Returns:    
        list: list of all finished books
    """
    if not annotations:
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZISFINISHED", 1)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZISFINISHED = 1"
    return db_utils.run_query(query)

def find_unfinished_books(annotations: bool = False) -> list:
    """
    Fetches all unfinished books

    Returns:
        list: list of all unfinished books
    """
    if not annotations:
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZISFINISHED", 0)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZISFINISHED = 0"
    return db_utils.run_query(query)

def find_explicit_books(annotations: bool = False) -> list:
    """
    Fetches all explicit books
    
    Returns:
        list: list of all explicit books
    """
    if not annotations:
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZISEXPLICIT", 1)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZISEXPLICIT = 1"
    return db_utils.run_query(query)

def find_locked_books(annotations: bool = False) -> list:
    """
    Fetches all locked books

    Returns:
        list: list of all locked books
    """
    if not annotations:
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZISLOCKED", 1)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZISLOCKED = 1"
    return db_utils.run_query(query)

def find_ephemeral_books(annotations: bool = False) -> list:
    """
    Fetches all ephemeral books

    Returns:
        list: list of all ephemeral books
    """
    if not annotations:
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZISEPHEMERAL", 1)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZISEPHEMERAL = 1"
    return db_utils.run_query(query)

def find_hidden_books(annotations: bool = False) -> list:
    """
    Fetches all hidden books

    Returns:
        list: list of all hidden books
    """
    if not annotations:
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZISHIDDEN", 1)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZISHIDDEN = 1"
    return db_utils.run_query(query)

def find_sample_books(annotations: bool = False) -> list:
    """
    Fetches all sample books
    
    Returns:
        list: list of all sample books
    """
    if not annotations:
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZISSAMPLE", 1)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZISSAMPLE = 1"
    return db_utils.run_query(query)

def find_store_audiobooks(annotations: bool = False) -> list:
    """
    Fetches all store audiobooks
    
    Returns:
        list: list of all store audiobooks
    """
    if not annotations:
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZISSTOREAUDIOBOOK", 1)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZISSTOREAUDIOBOOK = 1"
    return db_utils.run_query(query)

def find_by_rating(rating: int, annotations: bool = False) -> list:
    """
    Fetches all books with the specified rating
    
    Args:   
        rating (int): rating of the book

    Returns:
        list: list of all books with the specified rating
    """
    if not annotations:
        return db_utils.find_by_field(fields_str, query_utils.BOOK_TABLE_NAME, "ZRATING", rating)
    query = query_utils.get_book_annotation_query()
    query += f"WHERE {query_utils.BOOK_TABLE_NAME}.ZRATING = {rating}"
    return db_utils.run_query(query)

def find_by_collection_id(collection_id: str) -> list:
    """
    Fetches all books in a collection

    Args:
        collection_id (str): id of the collection

    Returns:
        list: list of all books in the collection
    """
    query = query_utils.book_collection_member_query()
    query += f"WHERE {query_utils.COLLECTIONMEMBER_TABLE_NAME}.ZCOLLECTION = {collection_id}"
    return db_utils.run_query(query)


from setuptools import setup, find_packages

setup(
    # name of the package -> must be the same as the name of the folder
    name='py_apple_books',
    version='0.1.0',
    description='Python library for Apple Books',
    # author details
    author='Vignesh Iyer',
    author_email='vgnshiyer@asu.edu',
    packages=find_packages(),
    license='MIT',
    # projects official homepage
    url='https://github.com/vgnshiyer/py-apple-books',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # license
        'License :: OSI Approved :: MIT License',

        # python versions supported
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],

    package_data={
        'py_apple_books': ['*.ini'],
    },
)
        


from dataclasses import dataclass, field
from py_apple_books.models.book import Book
from typing import List

@dataclass
class Collection:
    # Identifiers
    id: str
    title: str

    # Status
    is_deleted: bool
    is_hidden: bool

    # Collection details
    details: str
    books: List[Book] = field(default_factory=list)

    def add_book(self, book: Book) -> None:
        """
        Adds a book to the collection.

        Args:
            book (Book): book to add to the collection
        """
        self.books.append(book)

    def __hash__(self) -> int:
        """
        Returns a hash value for the Collection object.

        Returns:
            int: hash value for the Collection object
        """
        return hash(self.id)